@@ -32,6 +32,7 @@ from oslo_utils import excutils
 from oslo_utils import imageutils
 from oslo_utils import units
 import requests
+import tenacity
 
 from ironic_lib.common.i18n import _
 from ironic_lib import disk_partitioner
@@ -79,6 +80,8 @@ _PARTED_PRINT_RE = re.compile(r"^(\d+):([\d\.]+)MiB:"
 CONFIGDRIVE_LABEL = "config-2"
 MAX_CONFIG_DRIVE_SIZE_MB = 64
 
+GPT_SIZE_SECTORS = 33
+
 # Maximum disk size supported by MBR is 2TB (2 * 1024 * 1024 MB)
 MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR = 2097152
 
@@ -345,7 +348,7 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     if commit:
         # write to the disk
         dp.commit()
-        _trigger_device_rescan(dev)
+        trigger_device_rescan(dev)
     return part_dict
 
 
@@ -391,25 +394,32 @@ def qemu_img_info(path):
     return imageutils.QemuImgInfo(out)
 
 
+def _retry_on_res_temp_unavailable(exc):
+    if (isinstance(exc, processutils.ProcessExecutionError)
+            and ('Resource temporarily unavailable' in exc.stderr
+                 or 'Cannot allocate memory' in exc.stderr)):
+        return True
+    return False
+
+
+@tenacity.retry(
+    retry=tenacity.retry_if_exception(_retry_on_res_temp_unavailable),
+    stop=tenacity.stop_after_attempt(CONF.disk_utils.image_convert_attempts),
+    reraise=True)
 def convert_image(source, dest, out_format, run_as_root=False):
     """Convert image to other format."""
-    # TODO(dtantsur): use the retrying library
-    for attempt in range(CONF.disk_utils.image_convert_attempts):
-        cmd = ('qemu-img', 'convert', '-O', out_format, source, dest)
-        try:
-            utils.execute(*cmd, run_as_root=run_as_root,
-                          prlimit=_qemu_img_limits(),
-                          use_standard_locale=True)
-        except processutils.ProcessExecutionError as e:
-            if ('Resource temporarily unavailable' in e.stderr
-                    and attempt < CONF.disk_utils.image_convert_attempts - 1):
-                LOG.debug('Failed to convert image, retrying. Error: %s', e)
-                # Sync disk caches before the next attempt
-                utils.execute('sync')
-            else:
-                raise
-        else:
-            return
+    cmd = ('qemu-img', 'convert', '-O', out_format, source, dest)
+    try:
+        utils.execute(*cmd, run_as_root=run_as_root,
+                      prlimit=_qemu_img_limits(),
+                      use_standard_locale=True)
+    except processutils.ProcessExecutionError as e:
+        if ('Resource temporarily unavailable' in e.stderr
+            or 'Cannot allocate memory' in e.stderr):
+            LOG.debug('Failed to convert image, retrying. Error: %s', e)
+            # Sync disk caches before the next attempt
+            utils.execute('sync')
+        raise
 
 
 def populate_image(src, dst, conv_flags=None):
@@ -489,13 +499,24 @@ def destroy_disk_metadata(dev, node_uuid):
     # erasing partition data.
     # This is the same bug as
     # https://bugs.launchpad.net/ironic-python-agent/+bug/1737556
+
+    # Overwrite the Primary GPT, catch very small partitions (like EBRs)
     dd_device = 'of=%s' % dev
-    gpt_backup = get_dev_block_size(dev) - 33
-    dd_seek = 'seek=%i' % gpt_backup
-    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, 'count=33',
+    dd_count = 'count=%s' % GPT_SIZE_SECTORS
+    dev_size = get_dev_block_size(dev)
+    if dev_size < GPT_SIZE_SECTORS:
+        dd_count = 'count=%s' % dev_size
+    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, dd_count,
                   run_as_root=True, use_standard_locale=True)
-    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, 'count=33',
-                  dd_seek, run_as_root=True, use_standard_locale=True)
+
+    # Overwrite the Secondary GPT, do this only if there could be one
+    if dev_size > GPT_SIZE_SECTORS:
+        gpt_backup = dev_size - GPT_SIZE_SECTORS
+        dd_seek = 'seek=%i' % gpt_backup
+        dd_count = 'count=%s' % GPT_SIZE_SECTORS
+        utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, dd_count,
+                      dd_seek, run_as_root=True, use_standard_locale=True)
+
     # Go ahead and let sgdisk run as well.
     utils.execute('sgdisk', '-Z', dev, run_as_root=True,
                   use_standard_locale=True)
@@ -540,9 +561,11 @@ def _get_configdrive(configdrive, node_uuid, tempdir=None):
 
     try:
         data = io.BytesIO(base64.decode_as_bytes(data))
-    except TypeError:
-        error_msg = (_('Config drive for node %s is not base64 encoded '
-                       'or the content is malformed.') % node_uuid)
+    except Exception as exc:
+        error_msg = (_('Config drive for node %(node)s is not base64 encoded '
+                       'or the content is malformed. %(cls)s: %(err)s.')
+                     % {'node': node_uuid, 'err': exc,
+                        'cls': type(exc).__name__})
         if is_url:
             error_msg += _(' Downloaded from "%s".') % configdrive
         raise exception.InstanceDeployFailure(error_msg)
@@ -767,11 +790,8 @@ def _get_labelled_partition(device_path, label, node_uuid):
     :returns: block device file for partition if it exists; otherwise it
               returns None.
     """
+    partprobe(device_path)
     try:
-        utils.execute('partprobe', device_path, run_as_root=True,
-                      attempts=CONF.disk_utils.partprobe_attempts)
-
-        # lsblk command
         output, err = utils.execute('lsblk', '-Po', 'name,label', device_path,
                                     check_exit_code=[0, 1],
                                     use_standard_locale=True, run_as_root=True)
@@ -876,7 +896,50 @@ def fix_gpt_partition(device, node_uuid):
         raise exception.InstanceDeployFailure(msg)
 
 
-def _trigger_device_rescan(device):
+def udev_settle():
+    """Wait for the udev event queue to settle.
+
+    Wait for the udev event queue to settle to make sure all devices
+    are detected once the machine boots up.
+
+    :return: True on success, False otherwise.
+    """
+    LOG.debug('Waiting until udev event queue is empty')
+    try:
+        utils.execute('udevadm', 'settle')
+    except processutils.ProcessExecutionError as e:
+        LOG.warning('Something went wrong when waiting for udev '
+                    'to settle. Error: %s', e)
+        return False
+    else:
+        return True
+
+
+def partprobe(device, attempts=None):
+    """Probe partitions on the given device.
+
+    :param device: The block device containing paritions that is attempting
+                   to be updated.
+    :param attempts: Number of attempts to run partprobe, the default is read
+                     from the configuration.
+    :return: True on success, False otherwise.
+    """
+    if attempts is None:
+        attempts = CONF.disk_utils.partprobe_attempts
+
+    try:
+        utils.execute('partprobe', device, run_as_root=True, attempts=attempts)
+    except (processutils.UnknownArgumentError,
+            processutils.ProcessExecutionError, OSError) as e:
+        LOG.warning("Unable to probe for partitions on device %(device)s, "
+                    "the partitioning table may be broken. Error: %(error)s",
+                    {'device': device, 'error': e})
+        return False
+    else:
+        return True
+
+
+def trigger_device_rescan(device, attempts=None):
     """Sync and trigger device rescan.
 
     Disk partition performed via parted, when performed on a ramdisk
@@ -894,25 +957,25 @@ def _trigger_device_rescan(device):
 
     :param device: The block device containing paritions that is attempting
                    to be updated.
+    :param attempts: Number of attempts to run partprobe, the default is read
+                     from the configuration.
+    :return: True on success, False otherwise.
     """
-    # TODO(TheJulia): This helper method was broken out for re-use, and
-    # does not have an explicit test case for itself, and it should, but
-    # it's steps are fairly explicitly checked with mocks on execute in
-    # the partition code.
-    LOG.debug('Explicitly calling sync to force buffer/cache flush.')
+    LOG.debug('Explicitly calling sync to force buffer/cache flush')
     utils.execute('sync')
     # Make sure any additions to the partitioning are reflected in the
     # kernel.
-    LOG.debug('Waiting until udev event queue is empty')
-    utils.execute('udevadm', 'settle')
+    udev_settle()
+    partprobe(device, attempts=attempts)
     try:
-        utils.execute('partprobe', device, run_as_root=True,
-                      attempts=CONF.disk_utils.partprobe_attempts)
         # Also verify that the partitioning is correct now.
         utils.execute('sgdisk', '-v', device, run_as_root=True)
     except processutils.ProcessExecutionError as exc:
-        LOG.warning('Failed to verify partitioning after creating '
-                    'partition(s): %s', exc)
+        LOG.warning('Failed to verify partition tables on device %(dev)s: '
+                    '%(err)s', {'dev': device, 'err': exc})
+        return False
+    else:
+        return True
 
 
 def create_config_drive_partition(node_uuid, device, configdrive):
@@ -998,7 +1061,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                               'mkpart', 'primary', 'fat32', startlimit,
                               endlimit, run_as_root=True)
             # Trigger device rescan
-            _trigger_device_rescan(device)
+            trigger_device_rescan(device)
 
             upd_parts = set(part['number'] for part in list_partitions(device))
             new_part = set(upd_parts) - set(cur_parts)
@@ -1015,8 +1078,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
             else:
                 config_drive_part = '%s%s' % (device, new_part.pop())
 
-            LOG.debug('Waiting until udev event queue is empty')
-            utils.execute('udevadm', 'settle')
+            udev_settle()
 
             # NOTE(vsaienko): check that devise actually exists,
             # it is not handled by udevadm when using ISCSI, for more info see:
