@@ -22,7 +22,6 @@ from unittest import mock
 
 from oslo_concurrency import processutils
 from oslo_config import cfg
-from oslo_serialization import base64
 from oslo_utils import imageutils
 import requests
 
@@ -800,6 +799,58 @@ class DestroyMetaDataTestCase(base.IronicLibTestCase):
         disk_utils.destroy_disk_metadata(self.dev, self.node_uuid)
         mock_exec.assert_has_calls(expected_call)
 
+    def test_destroy_disk_metadata_ebr(self, mock_exec):
+        expected_calls = [mock.call('wipefs', '--force', '--all', 'fake-dev',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
+                          mock.call('blockdev', '--getsz', 'fake-dev',
+                                    check_exit_code=[0],
+                                    run_as_root=True),
+                          mock.call('dd', 'bs=512', 'if=/dev/zero',
+                                    'of=fake-dev', 'count=2',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
+                          mock.call('sgdisk', '-Z', 'fake-dev',
+                                    run_as_root=True,
+                                    use_standard_locale=True)]
+        mock_exec.side_effect = iter([
+            (None, None),
+            ('2\n', None),  # an EBR is 2 sectors
+            (None, None),
+            (None, None),
+            (None, None),
+            (None, None)])
+        disk_utils.destroy_disk_metadata(self.dev, self.node_uuid)
+        mock_exec.assert_has_calls(expected_calls)
+
+    def test_destroy_disk_metadata_tiny_partition(self, mock_exec):
+        expected_calls = [mock.call('wipefs', '--force', '--all', 'fake-dev',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
+                          mock.call('blockdev', '--getsz', 'fake-dev',
+                                    check_exit_code=[0],
+                                    run_as_root=True),
+                          mock.call('dd', 'bs=512', 'if=/dev/zero',
+                                    'of=fake-dev', 'count=33',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
+                          mock.call('dd', 'bs=512', 'if=/dev/zero',
+                                    'of=fake-dev', 'count=33', 'seek=9',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
+                          mock.call('sgdisk', '-Z', 'fake-dev',
+                                    run_as_root=True,
+                                    use_standard_locale=True)]
+        mock_exec.side_effect = iter([
+            (None, None),
+            ('42\n', None),
+            (None, None),
+            (None, None),
+            (None, None),
+            (None, None)])
+        disk_utils.destroy_disk_metadata(self.dev, self.node_uuid)
+        mock_exec.assert_has_calls(expected_calls)
+
 
 @mock.patch.object(utils, 'execute', autospec=True)
 class GetDeviceBlockSizeTestCase(base.IronicLibTestCase):
@@ -845,9 +896,9 @@ class PopulateImageTestCase(base.IronicLibTestCase):
         self.assertFalse(mock_dd.called)
 
 
-# NOTE(TheJulia): _trigger_device_rescan is systemwide thus pointless
+# NOTE(TheJulia): trigger_device_rescan is systemwide thus pointless
 # to execute in the file test case. Also, CI unit test jobs lack sgdisk.
-@mock.patch.object(disk_utils, '_trigger_device_rescan', lambda *_: None)
+@mock.patch.object(disk_utils, 'trigger_device_rescan', lambda *_: None)
 @mock.patch.object(utils, 'wait_for_disk_to_become_available', lambda *_: None)
 @mock.patch.object(disk_utils, 'is_block_device', lambda d: True)
 @mock.patch.object(disk_utils, 'block_uuid', lambda p: 'uuid')
@@ -961,14 +1012,10 @@ class GetConfigdriveTestCase(base.IronicLibTestCase):
                           'http://1.2.3.4/cd', 'fake-node-uuid')
         self.assertFalse(mock_copy.called)
 
-    @mock.patch.object(base64, 'decode_as_bytes', autospec=True)
-    def test_get_configdrive_base64_error(self, mock_b64, mock_requests,
-                                          mock_copy):
-        mock_b64.side_effect = TypeError
+    def test_get_configdrive_base64_error(self, mock_requests, mock_copy):
         self.assertRaises(exception.InstanceDeployFailure,
                           disk_utils._get_configdrive,
                           'malformed', 'fake-node-uuid')
-        mock_b64.assert_called_once_with('malformed')
         self.assertFalse(mock_copy.called)
 
     @mock.patch.object(gzip, 'GzipFile', autospec=True)
@@ -1049,10 +1096,8 @@ class OtherFunctionTestCase(base.IronicLibTestCase):
     def test_convert_image_retries(self, execute_mock):
         ret_err = 'qemu: qemu_thread_create: Resource temporarily unavailable'
         execute_mock.side_effect = [
-            processutils.ProcessExecutionError(stderr=ret_err),
-            ('', ''),
-            processutils.ProcessExecutionError(stderr=ret_err),
-            ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
             ('', ''),
         ]
 
@@ -1071,13 +1116,35 @@ class OtherFunctionTestCase(base.IronicLibTestCase):
         ])
 
     @mock.patch.object(utils, 'execute', autospec=True)
-    def test_convert_image_fails(self, execute_mock):
+    def test_convert_image_retries_alternate_error(self, execute_mock):
+        ret_err = 'Failed to allocate memory: Cannot allocate memory\n'
+        execute_mock.side_effect = [
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            ('', ''),
+        ]
+
+        disk_utils.convert_image('source', 'dest', 'out_format')
+        convert_call = mock.call('qemu-img', 'convert', '-O',
+                                 'out_format', 'source', 'dest',
+                                 run_as_root=False,
+                                 prlimit=mock.ANY,
+                                 use_standard_locale=True)
+        execute_mock.assert_has_calls([
+            convert_call,
+            mock.call('sync'),
+            convert_call,
+            mock.call('sync'),
+            convert_call,
+        ])
+
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_convert_image_retries_and_fails(self, execute_mock):
         ret_err = 'qemu: qemu_thread_create: Resource temporarily unavailable'
         execute_mock.side_effect = [
-            processutils.ProcessExecutionError(stderr=ret_err),
-            ('', ''),
-            processutils.ProcessExecutionError(stderr=ret_err),
-            ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
             processutils.ProcessExecutionError(stderr=ret_err),
         ]
 
@@ -1097,6 +1164,25 @@ class OtherFunctionTestCase(base.IronicLibTestCase):
             convert_call,
         ])
 
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_convert_image_just_fails(self, execute_mock):
+        ret_err = 'Aliens'
+        execute_mock.side_effect = [
+            processutils.ProcessExecutionError(stderr=ret_err),
+        ]
+
+        self.assertRaises(processutils.ProcessExecutionError,
+                          disk_utils.convert_image,
+                          'source', 'dest', 'out_format')
+        convert_call = mock.call('qemu-img', 'convert', '-O',
+                                 'out_format', 'source', 'dest',
+                                 run_as_root=False,
+                                 prlimit=mock.ANY,
+                                 use_standard_locale=True)
+        execute_mock.assert_has_calls([
+            convert_call,
+        ])
+
     @mock.patch.object(os.path, 'getsize', autospec=True)
     @mock.patch.object(disk_utils, 'qemu_img_info', autospec=True)
     def test_get_image_mb(self, mock_qinfo, mock_getsize):
@@ -1253,8 +1339,13 @@ class WholeDiskPartitionTestCases(base.IronicLibTestCase):
                                'Failed to retrieve partition labels',
                                disk_utils._get_labelled_partition, self.dev,
                                self.config_part_label, self.node_uuid)
-        mock_execute.assert_called_once_with(
-            'partprobe', self.dev, run_as_root=True, attempts=10)
+        execute_calls = [
+            mock.call('partprobe', self.dev, run_as_root=True, attempts=10),
+            mock.call('lsblk', '-Po', 'name,label', self.dev,
+                      check_exit_code=[0, 1],
+                      use_standard_locale=True, run_as_root=True)
+        ]
+        mock_execute.assert_has_calls(execute_calls)
         self.assertEqual(1, mock_log.call_count)
 
     def _test_is_disk_larger_than_max_size(self, mock_execute, blk_out):
@@ -1880,3 +1971,37 @@ class WholeDiskConfigDriveTestCases(base.IronicLibTestCase):
         mock_fix_gpt_partition.assert_called_with(self.dev, self.node_uuid)
         mock_is_disk_gpt.assert_called_with(self.dev, self.node_uuid)
         mock_count.assert_called_once_with(self.dev)
+
+
+@mock.patch.object(utils, 'execute', autospec=True)
+class TriggerDeviceRescanTestCase(base.IronicLibTestCase):
+    def test_trigger(self, mock_execute):
+        self.assertTrue(disk_utils.trigger_device_rescan('/dev/fake'))
+        mock_execute.assert_has_calls([
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', '/dev/fake', run_as_root=True, attempts=10),
+            mock.call('sgdisk', '-v', '/dev/fake', run_as_root=True),
+        ])
+
+    def test_custom_attempts(self, mock_execute):
+        self.assertTrue(
+            disk_utils.trigger_device_rescan('/dev/fake', attempts=1))
+        mock_execute.assert_has_calls([
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', '/dev/fake', run_as_root=True, attempts=1),
+            mock.call('sgdisk', '-v', '/dev/fake', run_as_root=True),
+        ])
+
+    def test_fails(self, mock_execute):
+        mock_execute.side_effect = [('', '')] * 3 + [
+            processutils.ProcessExecutionError
+        ]
+        self.assertFalse(disk_utils.trigger_device_rescan('/dev/fake'))
+        mock_execute.assert_has_calls([
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', '/dev/fake', run_as_root=True, attempts=10),
+            mock.call('sgdisk', '-v', '/dev/fake', run_as_root=True),
+        ])
